"""
Message context.

"""
from typing import Dict

from microcosm.api import defaults, typed
from microcosm.config.types import boolean
from microcosm_logging.decorators import logger

from microcosm_pubsub.constants import TTL_KEY, URI_KEY
from microcosm_pubsub.message import SQSMessage


@defaults(
    enable_ttl=typed(boolean, default_value=True),
    initial_ttl=typed(int, default_value=32),
)
@logger
class SQSMessageContext:
    """
    Factory for per-message contexts.

    """
    def __init__(self, graph):
        self.enable_ttl = graph.config.sqs_message_context.enable_ttl
        self.initial_ttl = graph.config.sqs_message_context.initial_ttl

    def __call__(self, context: SQSMessage, **kwargs) -> Dict[str, str]:
        """
        Create a new context from a message.

        """
        return self.from_sqs_message(context, **kwargs)

    def from_sqs_message(self, message: SQSMessage, **kwargs):
        context: Dict = dict(message.opaque_data)

        context.update(
            # include the message id
            message_id=message.message_id,
            **kwargs,
        )

        # include the TTL (if enabled)
        if self.enable_ttl:
            ttl = message.ttl if message.ttl is not None else self.initial_ttl
            context[TTL_KEY] = str(ttl - 1)

        # include the URI (if there is one)
        if message.uri:
            context[URI_KEY] = message.uri

        return context



"""
PubSub control flow.

"""


class TopicNotDefinedError(Exception):
    """
    No topic was configured for message media type.

    """
    pass


class Nack(Exception):
    """
    Retry the message after a configured number of seconds.

    Note that retry behavior is contingent on both the backoff policy
    and the SQS dead letter queue (DLQ) configuration, if any.

    """
    def __init__(self, visibility_timeout_seconds):
        self.visibility_timeout_seconds = visibility_timeout_seconds

    def __repr__(self):
        return "Nack({})".format(self.visibility_timeout_seconds)


class SkipMessage(Exception):
    """
    Stop processing a message and do not retry processing.

    Indicates that message content failed a precondition within a handler.

    """
    def __init__(self, reason, extra=None):
        super().__init__(reason)
        self.extra = extra or dict()


class IgnoreMessage(Exception):
    """
    Do not processing a message and do not count it for statistics.

    Indicates that message was not processable by the daemon.

    """
    def __init__(self, reason, extra=None):
        super().__init__(reason)
        self.extra = extra or dict()


class TTLExpired(Exception):
    """
    Control-flow exception to skip messages in infinte loop.

    """
    def __init__(self, reason=None, extra=None):
        super().__init__(reason)
        self.extra = extra or dict()



"""
case("value").then(...)

"""
from microcosm_pubsub.chain import Chain


class CaseStatement:
    def __init__(self, switch, key):
        self.switch = switch
        self.key = key

    def then(self, *args, **kwargs):
        self.switch._add_action_for_key(self.key, Chain.make(*args, **kwargs))
        return self.switch



"""
Message handling result.

"""
from dataclasses import dataclass, field
from enum import Enum, unique
from logging import DEBUG, INFO, WARNING
from sys import exc_info
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)

from microcosm_pubsub.errors import (
    IgnoreMessage,
    Nack,
    SkipMessage,
    TTLExpired,
)
from microcosm_pubsub.message import SQSMessage


@dataclass
class MessageHandlingResultTypeInfo:
    name: str
    level: int
    retry: bool = False
    exc_info: bool = False


@unique
class MessageHandlingResultType(Enum):
    # Messaging handling aborted due to too many attempts.
    EXPIRED = MessageHandlingResultTypeInfo(name="EXPIRED", level=WARNING)

    # Messaging handling failed.
    FAILED = MessageHandlingResultTypeInfo(name="FAILED", level=WARNING, exc_info=True, retry=True)

    # Message was not handled.
    IGNORED = MessageHandlingResultTypeInfo(name="IGNORED", level=DEBUG)

    # Messaging handling was intentionally retried.
    RETRIED = MessageHandlingResultTypeInfo(name="RETRIED", level=INFO, retry=True)

    # Message handling succeeded.
    SUCCEEDED = MessageHandlingResultTypeInfo(name="SUCCEEDED", level=INFO)

    # Message handling was skipped.
    # ("Upon closer inspection these are loafers")
    SKIPPED = MessageHandlingResultTypeInfo(name="SKIPPED", level=INFO)

    def __str__(self):
        return self.name

    @property
    def exc_info(self):
        return self.value.exc_info

    @property
    def level(self):
        return self.value.level

    @property
    def retry(self):
        return self.value.retry


@dataclass
class MessageHandlingResult:
    media_type: str
    result: MessageHandlingResultType
    exc_info: Optional[Tuple[Any, Any, Any]] = None
    extra: Dict[str, str] = field(default_factory=dict)
    elapsed_time: Optional[float] = None
    handle_start_time: Optional[float] = None
    retry_timeout_seconds: Optional[int] = None

    @classmethod
    def invoke(cls, handler, message: SQSMessage):
        try:
            success = handler(message.content)
            return cls.from_result(message, bool(success))
        except Exception as error:
            return cls.from_error(message, error)

    @classmethod
    def from_result(cls, message: SQSMessage, success: bool):
        if success:
            result = MessageHandlingResultType.SUCCEEDED
        else:
            result = MessageHandlingResultType.SKIPPED

        return cls(
            media_type=message.media_type,
            result=result,
        )

    @classmethod
    def from_error(cls, message: SQSMessage, error: Exception, **kwargs):
        if isinstance(error, IgnoreMessage):
            return cls(
                extra=error.extra,
                media_type=message.media_type,
                result=MessageHandlingResultType.IGNORED,
            )

        if isinstance(error, SkipMessage):
            return cls(
                extra=dict(
                    reason=str(error),
                    **error.extra
                ),
                media_type=message.media_type,
                result=MessageHandlingResultType.SKIPPED,
            )

        if isinstance(error, TTLExpired):
            return cls(
                extra=error.extra,
                media_type=message.media_type,
                result=MessageHandlingResultType.EXPIRED,
            )

        if isinstance(error, Nack):
            return cls(
                media_type=message.media_type,
                result=MessageHandlingResultType.RETRIED,
                retry_timeout_seconds=error.visibility_timeout_seconds,
            )

        return cls(
            exc_info=exc_info(),
            media_type=message.media_type,
            result=MessageHandlingResultType.FAILED,
        )

    def log(self, logger, opaque):
        """
        Log this result.

        """
        entry = f"Result for media type: {self.media_type} was : {self.result} "
        logger.log(
            self.result.level,
            entry,
            exc_info=self.exc_info,
            extra=dict(**self.extra, **opaque.as_dict()),
        )

    def resolve(self, message):
        if self.result.retry:
            message.nack(self.retry_timeout_seconds)
        else:
            message.ack()
        return self



from marshmallow import fields

from microcosm_pubsub.codecs import PubSubMessageSchema


class URIMessageSchema(PubSubMessageSchema):
    """
    Define a baseline message schema that points to (the URI of) a resource that experienced a lifecycle change.

    By convention, pubsub messages should be *references* to something that happened within some other
    source-of-truth (e.g. a CRUD microservice). Because there are inherent race conditions between publishing
    a message and commit a persistent transaction, message consumers are expected to call back to source-of-truth
    (e.g. via HTTP) and fetch the current resource value. In the event that message has not been committed yet,
    the consumer can retry a few times before giving up (e.g. via SQS dead-lettering).

    """
    def __init__(self, media_type, **kwargs):
        super().__init__(**kwargs)
        self.MEDIA_TYPE = media_type

    uri = fields.String(required=True)

    def deserialize_media_type(self, obj):
        return self.MEDIA_TYPE


class ChangedURIMessageSchema(URIMessageSchema):
    """
    Define a baseline message schema that points to the URI of a updated resource, with the updated value.

    By convention, pubsub messages are a reference to something which happened, but changed messages can
    be published (and handled) before the changes are comitted to DB. Keeping the changed field on the message
    allows us to retry handling the message in case value doesn't match the one in the message.

    """
    field_name = fields.String()
    new_value = fields.String()


class IdentityMessageSchema(PubSubMessageSchema):
    """
    Define a baseline message schema that points to (the identity of) a resource that experienced a lifecycle change.

    In general, these conventions prefer URI-based messages, but for some lifecycle changes (e.g. deletion),
    a URI may not be available.

    """
    def __init__(self, media_type, **kwargs):
        super().__init__(**kwargs)
        self.MEDIA_TYPE = media_type

    id = fields.String(required=True)

    def deserialize_media_type(self, obj):
        return self.MEDIA_TYPE



from dataclasses import dataclass
from json import loads
from typing import Iterable, Mapping, Optional
from unittest.mock import Mock

from hamcrest import not_none
from hamcrest.core.helpers.wrap_matcher import wrap_matcher
from hamcrest.library.object.hasproperty import IsObjectWithProperty


@dataclass
class PublishedMessage:
    """
    A published message.

    """
    topic_arn: str
    message: Mapping[str, str]

    @property
    def media_type(self) -> Optional[str]:
        return self.message.get("mediaType")

    @property
    def uri(self) -> Optional[str]:
        return self.message.get("uri")

    @classmethod
    def from_call(cls, call) -> Optional["PublishedMessage"]:
        """
        Extract a published message from a mock call.

        """
        # NB: _Call is a tuple of (name, args, kwargs)
        name, args, kwargs = call

        # Case 1: we've mocked the boto3 SNS client
        if "TopicArn" in kwargs and "Message" in kwargs:
            topic_arn = kwargs["TopicArn"]
            # NB: we could use the envelope functions to parse the inner message
            message = loads(kwargs["Message"])
            return cls(
                topic_arn=topic_arn,
                message=message,
            )

        # Case 2: we've mocked the entire SNSProducer
        if args and args[0].startswith("application/vnd.globality") and kwargs:
            topic_arn = None

            message = kwargs.copy()
            message.update(mediaType=args[0])

            return cls(
                topic_arn=topic_arn,
                message=message,
            )

        # calls with unexpected format will be filtered
        return None

    @staticmethod
    def iter_from_mock_calls(mock_calls) -> Iterable["PublishedMessage"]:
        """
        Iterate over published messages among mock calls.

        """
        return list(filter(None, (
            PublishedMessage.from_call(mock_call)
            for mock_call in mock_calls
        )))

    @staticmethod
    def iter_from_sns_producer(sns_producer) -> Iterable["PublishedMessage"]:
        """
        Iterate over published messages from a mocked SNSProducer.

        """
        if isinstance(sns_producer, Mock):
            target = sns_producer.produce
        else:
            target = sns_producer.sns_client.publish

        return PublishedMessage.iter_from_mock_calls(target.mock_calls)


class PublishedMessageMatcher(IsObjectWithProperty):
    """
    A matcher that validates a single published message property.

    This matcher is primarily a cosmetic change describe errors in terms of messages
    (and not generic objects).

    """
    def describe_to(self, description):
        return description.append_text(
            f"a message with a {self.property_name} matching ",
        ).append_description_of(
            self.value_matcher,
        )


def has_media_type(match=None):
    # NB: if no matcher is provider, just check that *some* media type exists
    if match is None:
        match = not_none()

    return PublishedMessageMatcher("media_type", wrap_matcher(match))


def has_uri(match=None):
    # NB: if no matcher is provider, just check that *some* uri exists
    if match is None:
        match = not_none()

    return PublishedMessageMatcher("uri", wrap_matcher(match))



"""
Matchers for message publishing.

"""
from typing import List

from hamcrest.core.helpers.wrap_matcher import wrap_matcher
from hamcrest.library.collection.issequence_containinginanyorder import (
    IsSequenceContainingInAnyOrder,
)
from hamcrest.library.collection.issequence_containinginorder import IsSequenceContainingInOrder

from microcosm_pubsub.matchers.message import PublishedMessage


def published_messages_for(sns_producer) -> List[PublishedMessage]:
    return list(PublishedMessage.iter_from_sns_producer(sns_producer))


class NonStrictPublishingMatcher(IsSequenceContainingInAnyOrder):
    """
    Matcher that validates a sequence of published messages extracted from an SNSProducer.

    """
    def matches(self, sns_producer, mismatch_description=None):
        return super().matches(
            published_messages_for(sns_producer),
            mismatch_description,
        )


class StrictPublishingMatcher(IsSequenceContainingInOrder):
    """
    Matcher that validates a sequence of published messages extracted from an SNSProducer.

    """
    def matches(self, sns_producer, mismatch_description=None):
        return super().matches(
            published_messages_for(sns_producer),
            mismatch_description,
        )


def published(*matchers, strict=True):
    return StrictPublishingMatcher([
        wrap_matcher(matcher)
        for matcher in matchers
    ])


def published_inanyorder(*matchers, strict=True):
    return NonStrictPublishingMatcher([
        wrap_matcher(matcher)
        for matcher in matchers
    ])


def published_nothing(*matchers):
    return StrictPublishingMatcher([])



"""
Message encoding and decoding.

"""
from json import dumps, loads

from marshmallow import (
    EXCLUDE,
    Schema,
    ValidationError,
    fields,
)


DEFAULT_MEDIA_TYPE = "application/json"


def enrich(error: ValidationError, schema: Schema):
    """
    Enrich a schema validation error with a human-interpretable
    description of a marshmallow schema.

    """
    messages = error.messages
    if isinstance(messages, dict):
        messages.update(dict(
            schema=str(schema),
        ))
    else:
        messages.append(
            str(schema),
        )

    return error


class PubSubMessageSchema(Schema):
    """
    Base schema for messages, including a media type.

    """
    media_type = fields.Method(
        # called by dump
        serialize="serialize_media_type",
        # called by load
        deserialize="deserialize_media_type",
        attribute="mediaType",
        # need to set missing to non-None or marshmallow won't call the deserialize function
        missing=DEFAULT_MEDIA_TYPE,
    )
    opaque_data = fields.Dict(
        attribute="opaqueData",
        required=False,
    )

    def serialize_media_type(self, message):
        """
        Fetch the media type from the message.

        """
        try:
            return message["mediaType"]
        except KeyError:
            raise ValidationError("Message did not define a media type")

    def deserialize_media_type(self, obj):
        """
        Return a custom media type.

        Should be overridden in subclasses.

        """
        return DEFAULT_MEDIA_TYPE


class MediaTypeSchema(Schema):
    """
    Custom schema for extracting media type.

    """
    mediaType = fields.String(required=True)


class PubSubMessageCodec:
    """
    Message encoder/decoder.

    """

    def __init__(self, schema):
        self.schema = schema

    def encode(self, dct=None, **kwargs):
        """
        Encode a message.

        Uses the appropriate codec to write JSON.

        """
        message = dct.copy() if dct else dict()
        message.update(kwargs)
        # We allow unknown fields to pass through (but not included here)
        # to accommodate unconditional parameter passing during produce()
        # e.g. passing in `uri` for IdentityMessages
        return dumps(self.schema.load(message, unknown=EXCLUDE))

    def decode(self, message):
        """
        Decode a message.

        Uses the appropriate codec to read JSON.

        """
        if not isinstance(message, dict):
            dct = loads(message)
        else:
            dct = message

        try:
            # load performs a validation and raises on error
            # Similarly, we exclude unknown fields to avoid errors on decode
            self.schema.load(dct, unknown=EXCLUDE)
        except ValidationError as error:
            # Add more useful information to logs for debugging validation errors.
            raise enrich(error, self.schema)

        return self.schema.dump(dct)



"""
when("foo").then(
    ...
).otherwise(
    ...
)

"""
from microcosm_pubsub.chain import Chain


class WhenStatement:
    """
    Condition chain control on a key.

    """
    def __init__(self, key):
        self.key = key
        self._then = None
        self._otherwise = None

    def then(self, *args, **kwargs):
        self._then = Chain.make(*args, **kwargs)
        return self

    def otherwise(self, *args, **kwargs):
        self._otherwise = Chain.make(*args, **kwargs)
        return self

    def __str__(self):
        return f"when_{self.key}"

    def __call__(self, context):
        if context[self.key]:
            if self._then:
                return self._then(context)
        elif self._otherwise:
            return self._otherwise(context)


def when(key):
    """
    Run one of two chains - based on a condition

    Example: when("arg")
        .then(Chain(...))
        .otherwise(Chain(...))

    """

    return WhenStatement(key)



"""
Parse the SQS envelope.

SQS defines an envelope with a message id, receipt handle, body, and so forth,
with the underlying message embedded in the body. There are multiple ways to
process this envelope, depending on the degree of validation and metadata desired.

"""
from abc import ABCMeta, abstractmethod
from hashlib import md5
from json import loads
from uuid import uuid4

from microcosm.api import defaults

from microcosm_pubsub.codecs import MediaTypeSchema, PubSubMessageCodec
from microcosm_pubsub.message import SQSMessage


class MessageBodyParser(metaclass=ABCMeta):
    """
    Mixin for parsing the data from an SQS message.

    """
    @abstractmethod
    def parse_message(self, body):
        """
        Extract the user-space portions of the message from the message body.

        :returns: a dictionary

        """
        pass


class MediaTypeAndContentParser(metaclass=ABCMeta):
    """
    Mixin for parsing a media type and content from a message body.

    """
    @abstractmethod
    def parse_media_type_and_content(self, message):
        """
        Extract the media type and content dictionary for a message body.

        :returns: a media type (string), and content (dict) tuple

        """
        pass


class RawMessageBodyParser(MessageBodyParser):

    def parse_message(self, body):
        """
        Assume the top-level message *IS* the message body.

        """
        return body


class RawMediaTypeAndContentParser(MediaTypeAndContentParser):

    def parse_media_type_and_content(self, message):
        content = message
        media_type = "application/json"
        return media_type, loads(content)


class NaiveMediaTypeAndContentParser(MediaTypeAndContentParser):

    def parse_media_type_and_content(self, message):
        media_type = message["mediaType"]
        content = {
            key: value
            for key, value in message.items()
            if key != "mediaType"
        }
        return media_type, content


class SNSMessageBodyParser(MessageBodyParser):

    def parse_message(self, body):
        """
        Extract the user-space portions of the message from the message body.

        When an SQS queue subscribes to an SNS topic, the user-space message is packaged up
        as JSON within the `Message` key of the top-level envelope.

        """
        body_dict = loads(body)
        try:
            return body_dict["Message"]
        except KeyError:
            # We're handling a raw message
            return body_dict


class CodecMediaTypeAndContentParser(MediaTypeAndContentParser):

    def __init__(self, graph):
        super(CodecMediaTypeAndContentParser, self).__init__(graph)
        self.media_type_codec = PubSubMessageCodec(MediaTypeSchema())
        self.pubsub_message_schema_registry = graph.pubsub_message_schema_registry

    def parse_media_type_and_content(self, message):
        """
        Decode the message once to extract its media type and then again with the correct codec.

        """
        base_message = self.media_type_codec.decode(message)
        media_type = base_message["mediaType"]
        try:
            content = self.pubsub_message_schema_registry.find(media_type).decode(message)
        except KeyError:
            return media_type, None
        else:
            return media_type, content


class SQSEnvelope(MessageBodyParser, MediaTypeAndContentParser):
    """
    Enveloping base class.

    """
    def __init__(self, graph):
        self.should_validate_md5 = graph.config.sqs_envelope.validate_md5

    def parse_raw_message(self, consumer, raw_message):
        """
        Create an `SQSMessage` from SQS data.

        """
        message_id = self.parse_message_id(raw_message)
        receipt_handle = self.parse_receipt_handle(raw_message)
        attributes = raw_message.get("Attributes", {})
        approximate_receive_count = int(attributes.get("ApproximateReceiveCount", 1))

        body = self.parse_body(raw_message)

        if self.should_validate_md5:
            self.validate_md5(raw_message, body)

        message = self.parse_message(body)
        media_type, content = self.parse_media_type_and_content(message)

        return SQSMessage(
            consumer=consumer,
            content=content,
            media_type=media_type,
            message_id=message_id,
            receipt_handle=receipt_handle,
            approximate_receive_count=approximate_receive_count,
        )

    def parse_message_id(self, raw_message):
        return raw_message["MessageId"]

    def parse_receipt_handle(self, raw_message):
        return raw_message["ReceiptHandle"]

    def parse_body(self, raw_message):
        return raw_message["Body"]

    def validate_md5(self, raw_message, body):
        """
        Validate the message body.

        Just checks for tampering; schema validation occurs once we know the type of message.

        """
        expected_md5_of_body = raw_message["MD5OfBody"]
        actual_md5_of_body = md5(body).hexdigest()
        if expected_md5_of_body != actual_md5_of_body:
            raise Exception("MD5 validation failed. Expected: {} Actual: {}".format(
                expected_md5_of_body,
                actual_md5_of_body,
            ))


class RawSQSEnvelope(RawMessageBodyParser, RawMediaTypeAndContentParser, SQSEnvelope):
    """
    Enveloping strategy that just passes raw JSON.

    """
    pass


class CodecSQSEnvelope(SNSMessageBodyParser, CodecMediaTypeAndContentParser, SQSEnvelope):
    """
    Enveloping strategy that uses a media type-driven message codec.

    """
    pass


class NaiveSQSEnvelope(RawMessageBodyParser, NaiveMediaTypeAndContentParser, SQSEnvelope):
    """
    Enveloping strategy for naive debugging.

    """
    def parse_message_id(self, raw_message):
        return "message-id-{}".format(uuid4())

    def parse_receipt_handle(self, raw_message):
        return "receipt-handle-{}".format(uuid4())

    def parse_body(self, raw_message):
        return raw_message


@defaults(
    strategy_name="CodecSQSEnvelope",
    validate_md5=False,
)
def configure_sqs_envelope(graph):
    strategy_name = graph.config.sqs_envelope.strategy_name
    # It should be possible to inject custom enveloping strategies as long as the
    # subclass is imported before the graph is initialized; alternative, this block
    # can be switched to using setuptools entry points.
    strategies = {
        strategy.__name__: strategy
        for strategy in SQSEnvelope.__subclasses__()
    }
    try:
        strategy = strategies[strategy_name]
    except KeyError:
        raise Exception("Unknown SQS enveloping strategy: {}".format(
            strategy_name,
        ))
    else:
        return strategy(graph)



"""
Enumeration of common lifecycle changes.

"""


class LifecycleChange(set):
    """
    A collection of CRUD lifecycle transitions that may be announced via pubsub.

    Note that `Changed` is strongly discouraged; use immutable resources instead.

    """
    Changed = "changed"
    Created = "created"
    Deleted = "deleted"

    def __init__(self, graph):
        super().__init__([
            LifecycleChange.Changed,
            LifecycleChange.Created,
            LifecycleChange.Deleted,
        ])

    def matches(self, media_type):
        parts = media_type.split(".")
        return any(
            item in parts
            for item in self
        )



"""
Custom fields.

"""
from uuid import UUID

from marshmallow.fields import Field, ValidationError


class UUIDField(Field):
    """
    UUID valued field, as either a string or UUID object.

    """
    def _serialize(self, value, attr, obj, **kwargs):
        """
        Serialize value as a uuid, either as a string or an UUID object.
        """
        if value is None:
            return None

        if isinstance(value, UUID):
            return str(value)
        else:
            return value

    def _deserialize(self, value, attr, data, **kwargs):
        """
        Deserialize value as a uuid.

        Handle both string and UUID object.
        """
        if value is None:
            return None

        try:
            if isinstance(value, UUID):
                return str(value)
            else:
                return str(UUID(value))

        except ValueError:
            raise ValidationError('Not a valid UUID.')



from abc import ABCMeta, abstractmethod

from microcosm_pubsub.chain import Chain
from microcosm_pubsub.handlers.uri_handler import URIHandler


class ChainHandler(metaclass=ABCMeta):
    """
    Resolve a chain on call. Pass to the chain the message to the chain.

    """
    @abstractmethod
    def get_chain(self):
        pass

    def __call__(self, message):
        return Chain(self.get_chain())(message=message)


class ChainURIHandler(URIHandler, metaclass=ABCMeta):
    """
    Base handler for URI-driven events based on URIHandler
    Resolve a chain on handle.
    Pass to the chain the message and the fetched resource.

    """
    @abstractmethod
    def get_chain(self):
        pass

    @property
    def resource_name(self):
        return "resource"

    def handle(self, message, uri, resource):
        kwargs = dict(
            message=message,
            uri=uri,
        )
        kwargs[self.resource_name] = resource
        chain = self.get_chain()
        chain(**kwargs)

        return True



from marshmallow import Schema, fields

from microcosm_pubsub.codecs import PubSubMessageSchema
from microcosm_pubsub.conventions import created
from microcosm_pubsub.decorators import schema


class BatchedMessageSchema(Schema):
    """
    A wrapper for a single message that is to be published in a
    MessageBatchSchema.

    """
    media_type = fields.String(required=True)
    message = fields.Raw(required=True)
    topic_arn = fields.String(required=True)
    opaque_data = fields.Raw(required=True)
    message_attributes = fields.Raw(required=True)


@schema
class MessageBatchSchema(PubSubMessageSchema):
    """
    A message indicating that a batch of messages needs to be published.

    """
    MEDIA_TYPE = created("batch_message")

    messages = fields.List(fields.Nested(BatchedMessageSchema), required=True)

    def deserialize_media_type(self, obj):
        return MessageBatchSchema.MEDIA_TYPE



"""
PubSub CLI

"""
from __future__ import print_function

from argparse import ArgumentParser
from json import dumps
from sys import stdout

from microcosm.api import create_object_graph
from microcosm.loaders import load_from_dict

from microcosm_pubsub.conventions import changed, created


def main():
    args = parse_args()
    args.func(args)


def parse_args():
    parser = ArgumentParser()

    # common arguments
    parser.add_argument(
        "--debug",
        action="store_true",
    )
    parser.add_argument(
        "--profile",
    )
    parser.add_argument(
        "--region",
    )

    subparsers = parser.add_subparsers()
    write_parser = subparsers.add_parser("write")
    write_parser.add_argument(
        "--topic-arn",
        required=True,
    )
    write_parser.add_argument(
        "--uri",
        required=True,
    )
    write_parser.add_argument(
        "--media-type",
        required=True,
    )
    write_parser.set_defaults(func=write)

    read_parser = subparsers.add_parser("read")
    read_parser.add_argument(
        "--queue-url",
        required=True,
    )
    read_parser.add_argument(
        "--nack",
        action="store_true",
    )
    read_parser.add_argument(
        "--nack-timeout",
        default=1,
        type=int,
    )
    read_parser.set_defaults(func=read)
    return parser.parse_args()


def write(args):
    loader = load_from_dict(
        sns_producer=dict(
            profile_name=args.profile,
            region_name=args.region,
        ),
        sns_topic_arns=dict(
            default=args.topic_arn,
        ),
    )

    graph = create_object_graph("pubsub", debug=args.debug, loader=loader)
    graph.use("pubsub_message_schema_registry")
    graph.use("sns_topic_arns")
    graph.use("sns_producer")
    graph.lock()

    if args.media_type.startswith("application"):
        media_type = args.media_type
    else:
        media_type = created(args.media_type)

    message_id = graph.sns_producer.produce(
        media_type=media_type,
        uri=args.uri,
    )
    print("Wrote SNS message: {}".format(message_id))  # noqa


def read(args):
    loader = load_from_dict(
        sqs_consumer=dict(
            profile_name=args.profile,
            region_name=args.region,
            sqs_queue_url=args.queue_url,
        ),
    )
    graph = create_object_graph("pubsub", debug=args.debug, loader=loader)
    graph.use("sqs_consumer")
    graph.lock()

    for message in graph.sqs_consumer.consume():
        print("Read SQS message: {} ({})".format(  # noqa
            message.message_id,
            message.approximate_receive_count,
        ))
        if args.nack:
            message.nack(args.nack_timeout)
        else:
            message.ack()


def make_naive_message():
    parser = ArgumentParser()

    parser.add_argument(
        "--action",
        choices=["created", "changed"],
    )
    parser.add_argument(
        "--resource-name",
    )
    parser.add_argument(
        "--uri",
    )
    args = parser.parse_args()

    action = dict(created=created, changed=changed)[args.action]

    stdout.write(dumps(dict(mediaType=action(args.resource_name), uri=args.uri)))
